{{#hasAnnotations}}
<div>
  <ol class="annotation-list">
    {{#annotations}}
    <li>{{{linkedNameWithParameters}}}</li>
    {{/annotations}}
  </ol>
</div>
{{/hasAnnotations}}
{{>name_summary}}{{{genericParameters}}} = {{{modelType.linkedName}}}</span>
