{{>head}}

  <div
      id="dartdoc-main-content"
      class="main-content"
      data-above-sidebar="{{ aboveSidebarPath }}"
      data-below-sidebar="{{ belowSidebarPath }}">
    {{#self}}
      <div>{{>source_link}}<h1><span class="kind-constructor">{{{nameWithGenerics}}}</span> {{kind}} {{>tags}}</h1></div>
    {{/self}}

    {{#constructor}}
    <section class="multi-line-signature">
      {{ >annotations }}
      {{#isConst}}const{{/isConst}}
      <span class="name {{#isDeprecated}}deprecated{{/isDeprecated}}">{{{nameWithGenerics}}}</span>(<wbr>{{#hasParameters}}{{{linkedParamsLines}}}{{/hasParameters}})
    </section>

    {{>documentation}}

    {{>source_code}}

    {{/constructor}}
  </div> <!-- /.main-content -->

  <div id="dartdoc-sidebar-left" class="sidebar sidebar-offcanvas-left">
    {{>search_sidebar}}
    <h5>{{parent.name}} {{parent.kind}}</h5>
    <div id="dartdoc-sidebar-left-content"></div>
  </div><!--/.sidebar-offcanvas-left-->

  <div id="dartdoc-sidebar-right" class="sidebar sidebar-offcanvas-right">
  </div><!--/.sidebar-offcanvas-->

{{>footer}}
