{{ #hasAvailableInstanceMethods }}
  <section
      class="summary offset-anchor{{ #publicInheritedInstanceMethods }} inherited{{ /publicInheritedInstanceMethods }}"
      id="instance-methods">
    <h2>Methods</h2>
    <dl class="callables">
      {{ #availableInstanceMethodsSorted }}
        {{ >callable }}
      {{ /availableInstanceMethodsSorted }}
    </dl>
  </section>
{{ /hasAvailableInstanceMethods }}