{{ #hasPublicConstructors }}
  <section class="summary offset-anchor" id="constructors">
    <h2>Constructors</h2>

    <dl class="constructor-summary-list">
      {{ #publicConstructorsSorted }}
        <dt id="{{htmlId}}" class="callable">
          <span class="name">{{{ linkedName }}}</span><span class="signature">({{{ linkedParams }}})</span>
        </dt>
        <dd>
          {{{ oneLineDoc }}}
          {{ #isConst }}
            <div class="constructor-modifier features">const</div>
          {{ /isConst }}
          {{ #isFactory }}
            <div class="constructor-modifier features">factory</div>
          {{ /isFactory }}
        </dd>
      {{ /publicConstructorsSorted }}
    </dl>
  </section>
{{ /hasPublicConstructors }}