{{ #hasPublicStaticMethods }}
  <section class="summary offset-anchor" id="static-methods">
    <h2>Static Methods</h2>
    <dl class="callables">
      {{#publicStaticMethodsSorted}}
        {{>callable}}
      {{/publicStaticMethodsSorted}}
    </dl>
  </section>
{{ /hasPublicStaticMethods }}